"""
Define all the env variable names that users
can use
"""

TRACKING_HOST = "MLF_HOST"
API_KEY = "MLF_API_KEY"

## NOTE: Global means this API key will be
# shared between servicefoundry and mlfoundry
TRACKING_HOST_GLOBAL = "TFY_HOST"
API_KEY_GLOBAL = "TFY_API_KEY"
##

TRACKING_TOKEN = "MLFLOW_TRACKING_TOKEN"
